"""This file contains checksum implementations."""

from typing import List


class BadCharacterException(ValueError):
    """An invalid character was found in your input."""


class BadAlphabetException(Exception):
    """There was a problem with the provided alphabet."""


class Damm32:
    """Implementation of the Damm algorithm."""

    BASE_SIZE = 32
    BIT_MASK = 37  # Bitmask from Table of Low-Weight Binary Irreducible Polynomials
    _DEFAULT_ALPHABET = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '2', '3', '4', '5', '6', '7']  # noqa: E501

    def __init__(self, alphabet: List[str] = _DEFAULT_ALPHABET) -> None:
        self._alphabet = alphabet
        self._check_alphabet_valid()

    def calculate(self, word: str) -> str:
        """
        Calculate the Damm check digit for a string.

        The string must consist only of characters in the alphabet.

        :param word: String to calculatee Damm check digit for.
        :returns: Single check digit character.
        """
        digits = self._to_digits(word)
        checkdigit = self._calculate_from_digits(digits)
        return self._alphabet[checkdigit]

    def verify(self, word: str) -> bool:
        """
        Verify that a string contains a valid Damm check digit.

        :param word: Word to check validity.
        :returns: Result of the check.
        """
        return self.calculate(word) == self._alphabet[0]

    def _check_alphabet_valid(self) -> None:
        """
        Check that the alphabet is valid.

        :raises BadAlphabetException: The alphabet was invalid.
        :raises BadCharacterException: At least one character in the alphabet is invalid.
        """
        length = len(self._alphabet)
        if length != self.BASE_SIZE:
            raise BadAlphabetException(
                f"Expected alphabet of length {self.BASE_SIZE}, got {length}",
            )

        length_unique = len(set(self._alphabet))
        if length_unique != self.BASE_SIZE:
            raise BadAlphabetException(
                f"Expected {self.BASE_SIZE} unique characters in "
                f"alphabet, got {length_unique }",
            )

        if not all(len(i) == 1 for i in self._alphabet):
            raise BadCharacterException(
                "Some characters in the provided alphabet were too long.",
            )

    def _calculate_from_digits(self, digits: List[int]) -> int:
        """
        Calculate the Damm check digit from a list of integer digits.

        :param digits: digits to calculate check digit from.
        :returns: The Damm check digit.
        """
        checkdigit = 0
        for digit in digits:
            checkdigit ^= digit
            checkdigit <<= 1
            if checkdigit >= self.BASE_SIZE:
                checkdigit ^= self.BIT_MASK
        return checkdigit

    def _to_digits(self, word: str) -> List[int]:
        """
        Convert a string to a list of integer digits.

        :param word: Word to convert.
        :returns: List of integer digits.
        :raises TypeError: Input is not a string.
        :raises BadCharacterException: Word contains an invalid character.
        """
        if type(word) is not str:
            raise TypeError("Input is not a string.")
        word = word.upper()
        try:
            return [self._alphabet.index(letter) for letter in word]
        except ValueError:
            raise BadCharacterException(f"Invalid character in {word}") from None

    def _to_word(self, digits: List[int]) -> str:
        """
        Convert a list of digits to a string.

        :param digits: digits to convert.
        :returns: constructed string.
        """
        return "".join([self._alphabet[digit] for digit in digits])



"""
Damm32 Library.

A python implementation of the Damm algorithm, in base 32.

"""

from .checksum import BadAlphabetException, BadCharacterException, Damm32

__all__ = ["BadAlphabetException", "BadCharacterException", "Damm32"]

__version__ = "1.2.1"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['damm32']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'damm32',
    'version': '1.2.1',
    'description': 'A pure-python implementation of the Damm Algorithm in Base 32.',
    'long_description': '# Damm32\n\n[![Tests](https://github.com/pyinv/damm32/actions/workflows/test.yml/badge.svg?branch=main)](https://github.com/pyinv/damm32/actions/workflows/test.yml)\n\nPython implementation of the Damm Algorithm in Base 32\n\nBy default, it uses an alphabet as specified in [RFC 4648](https://tools.ietf.org/html/rfc4648) which contains 32 alphanumeric characters, with similar looking characters removed. The padding symbol is not included.\n\n## Installation\n\nThe package is available on [PyPI](https://pypi.org/project/damm32/) and can be installed using pip: `pip install damm32`\n\nIt is also available on the [Arch User Repository](https://aur.archlinux.org/packages/python-damm32/) as `python-damm32`.\n\nAlternatively, you can clone the repository and use the module.\n\n## Usage\n\nThe module contains a single class called `Damm32`, this class implements the methods for the checksum.\n\n```\nfrom damm32 import Damm32\n\nd32 = Damm32()\n\ndigit = d32.calculate("HELLO")\n\nd32.verify("HELLO" + digit)\n\n```\n\nYou can also pass an list of length 32 to the constructor for the class to specify your alphabet.\n\n```\nfrom damm32 import Damm32\n\nd32 = Damm32([\'A\', \'B\', \'C\', \'D\', \'E\', \'F\', \'G\', \'H\', \'I\', \'J\', \'K\', \'L\', \'M\', \'N\', \'O\', \'P\', \'Q\', \'R\', \'S\', \'T\', \'U\', \'V\', \'W\', \'X\', \'Y\', \'Z\', \'2\', \'3\', \'4\', \'5\', \'6\', \'7\'])\n\n```\n\n## How it works\n\nThis is an implementation of the [Damm Algorithm](https://en.wikipedia.org/wiki/Damm_algorithm) for use in Base 32 systems.\n\nIt will detect all occurrences of the two most frequently appearing types of transcription errors, namely altering one single digit, and transposing two adjacent digits (including the transposition of the trailing check digit and the preceding digit).\n\nSince prepending leading zeros does not affect the check digit, variable length codes should not be verified together since, e.g., 0, 01, and 001, etc. produce the same check digit. However, all checksum algorithms are vulnerable to this.\n\nThe implementation uses a bitmask from [Table of Low-Weight Binary Irreducible Polynomials](https://www.hpl.hp.com/techreports/98/HPL-98-135.pdf) to enable calculation of the check digit without constructing the quasigroup.',
    'author': 'Dan Trickey',
    'author_email': 'dan@trickey.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pyinv/damm32',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
